from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from clowncar.server import Server


class Error(Exception):
    pass


class BackendError(Error):
    def __init__(self, message, server):
        # type: (str, Server) -> None
        self.message = message
        self.server = server

    def __str__(self):
        # type: () -> str
        return "({}:{}) - {}".format(self.server.hostname, self.server.port, self.message)


class BackendConnectionError(BackendError):
    pass


class BackendIntegrityError(BackendError):
    pass


class TimeTravelNotAllowed(BackendError):
    pass


class BackendMaxDriftError(BackendError):
    pass


class ResourceError(Error):
    pass


class ResourceNotFound(ResourceError):
    pass



from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from typing import Any, Dict, Text


@pytest.fixture
def service_account_response(request):
    # type: (str) -> Dict[Text, Any]
    return {
        u"checkpoint": 10,
        u"checkpoint_time": 1000,
        u"data": {
            u"groups": {},
            u"permissions": [
                {u"argument": u"shell", u"granted_on": 1452796706.894347, u"permission": u"sudo"}
            ],
            u"user": {
                u"enabled": True,
                u"metadata": [],
                u"name": u"service@a.co",
                u"public_keys": [],
                u"role_user": False,
                u"service_account": {
                    u"description": u"Some service account",
                    u"machine_set": u"shell",
                    u"owner": u"security-team",
                },
            },
        },
        u"status": u"ok",
    }


@pytest.fixture
def permission_response(request):
    # type: (str) -> Dict[Text, Any]
    return {
        u"checkpoint": 3,
        u"checkpoint_time": 1605842894,
        u"data": {
            u"permission": {u"name": "grouper.audit.security"},
            u"groups": {},
            u"service_accounts": {},
            u"audited": False,
        },
        u"status": u"ok",
    }


@pytest.fixture
def user_response(request):
    # type: (str) -> Dict[Text, Any]
    return {
        u"checkpoint": 10,
        u"checkpoint_time": 1000,
        u"data": {
            u"groups": {
                u"all-teams": {
                    u"distance": 3,
                    u"name": u"all-teams",
                    u"path": [u"oliver@a.co", u"security-team", u"team-infra", u"all-teams"],
                    u"role": 0,
                    u"rolename": u"member",
                },
                u"sad-team": {
                    u"distance": 1,
                    u"name": u"sad-team",
                    u"path": [u"oliver@a.co", u"sad-team"],
                    u"role": 0,
                    u"rolename": u"member",
                },
                u"security-team": {
                    u"distance": 1,
                    u"name": u"security-team",
                    u"path": [u"oliver@a.co", u"security-team"],
                    u"role": 2,
                    u"rolename": u"owner",
                },
                u"team-infra": {
                    u"distance": 2,
                    u"name": u"team-infra",
                    u"path": [u"oliver@a.co", u"security-team", u"team-infra"],
                    u"role": 0,
                    u"rolename": u"member",
                },
            },
            u"permissions": [
                {
                    u"argument": u"shell",
                    u"distance": 2,
                    u"granted_on": 1452796706.894347,
                    u"path": [u"oliver@a.co", u"security-team", u"team-infra"],
                    u"permission": u"sudo",
                }
            ],
            u"user": {
                u"enabled": True,
                u"metadata": [],
                u"name": u"oliver@a.co",
                u"public_keys": [],
                u"role_user": False,
            },
        },
        u"status": u"ok",
    }



from typing import TYPE_CHECKING

from six import iteritems, string_types

if TYPE_CHECKING:
    from typing import Any, Dict, List, Optional, Union


class ResourceDict(dict):
    def __call__(self, direct=False, roles=None):
        # type: (bool, Union[str, List[str]]) -> Dict[str, Any]
        if isinstance(roles, string_types):
            roles = [roles]
        new_dict = {}
        for key, value in iteritems(self):
            if direct and value.get("distance", 0) != 1:
                continue
            if roles and value.get("rolename") not in roles:
                continue
            new_dict[key] = value
        return new_dict


class Group(object):
    def __init__(
        self,
        groups,  # type: Dict[str, Dict[str, Any]]
        users,  # type: Dict[str, Dict[str, Any]]
        subgroups,  # type: Dict[str, Dict[str, Any]]
        permissions,  # type: List[Dict[str, Any]]
        audited,  # type: bool
        contacts,  # type: Dict[str, str]
    ):
        # type: (...) -> None
        self.groups = ResourceDict(groups)
        self.users = ResourceDict(users)
        self.subgroups = ResourceDict(subgroups)
        self.permissions = [
            MappedPermission.from_payload(permission) for permission in permissions
        ]
        self.audited = audited
        self.contacts = contacts

    @classmethod
    def from_payload(cls, payload):
        # type: (Dict[str, Any]) -> Group
        return cls(
            payload["data"]["groups"],
            payload["data"]["users"],
            payload["data"]["subgroups"],
            payload["data"]["permissions"],
            # New values may not exist in the JSON objects, so we need to be
            # careful.
            payload["data"].get("audited", False),
            # TODO(lfaraone): Figure out why we don't always return 'group'
            payload["data"].get("group", dict()).get("contacts", dict()),
        )


class User(object):
    def __init__(
        self,
        groups,  # type: Dict[str, Dict[str, Any]]
        public_keys,  # type: List[Dict[str, Any]]
        permissions,  # type: List[Dict[str, Any]]
        metadata,  # type: List[Dict[str, str]]
        enabled,  # type: bool
        passwords,  # type: List[Dict[str, str]]
        service_account,  # type: Optional[Dict[str, str]]
    ):
        # type: (...) -> None
        self.groups = ResourceDict(groups)
        self.passwords = passwords
        self.public_keys = public_keys
        self.enabled = enabled
        self.service_account = service_account
        self.permissions = [
            MappedPermission.from_payload(permission) for permission in permissions
        ]
        self.metadata = {md["data_key"]: UserMetadata.from_payload(md) for md in metadata}

    @classmethod
    def from_payload(cls, payload):
        # type: (Dict[str, Any]) -> User
        return cls(
            payload["data"]["groups"],
            payload["data"]["user"]["public_keys"],
            payload["data"]["permissions"],
            payload["data"]["user"]["metadata"],
            payload["data"]["user"]["enabled"],
            # New values may not exist in the JSON objects, so we need to be careful.
            payload["data"]["user"].get("passwords", []),
            # Optional field only present for service accounts.
            #
            # TODO(rra): ServiceAccount objects should lift these up to top-level properties and
            # User objects should not have this, and we should return a ServiceAccount when
            # retrieving a User with this set.
            payload["data"]["user"].get("service_account"),
        )


class ServiceAccount(User):
    pass


class Permission(object):
    def __init__(self, groups, audited):
        # type: (Dict[str, Dict[str, Any]], bool) -> None
        self.groups = {
            groupname: Group.from_payload({"data": groups[groupname]}) for groupname in groups
        }
        self.audited = audited

    @classmethod
    def from_payload(cls, payload):
        # type: (Dict[str, Any]) -> Permission
        return cls(payload["data"]["groups"], payload["data"]["audited"])


class MappedPermission(object):
    def __init__(self, permission, argument, granted_on, distance, path, audited):
        # type: (str, str, float, Optional[int], Optional[List[str]], Optional[bool]) -> None
        self.permission = permission
        self.argument = argument
        self.granted_on = granted_on
        self.distance = distance
        self.path = path
        self.audited = audited

    @classmethod
    def from_payload(cls, payload):
        # type: (Dict[str, Any]) -> MappedPermission
        return cls(
            payload["permission"],
            payload["argument"],
            payload["granted_on"],
            payload.get("distance"),
            payload.get("path"),
            payload.get("audited"),
        )


class UserMetadata(object):
    def __init__(self, key, value, last_modified):
        # type: (str, str, str) -> None
        self.key = key
        self.value = value
        self.last_modified = last_modified

    @classmethod
    def from_payload(cls, payload):
        # type: (Dict[str, Any]) -> UserMetadata
        return cls(payload["data_key"], payload["data_value"], payload["last_modified"])



import time
from threading import Lock
from typing import TYPE_CHECKING

from clowncar.server import Server

if TYPE_CHECKING:
    from typing import Set


class ServersFileLoader(object):
    def __init__(self, filename, cache_timeout=300):
        # type: (str, int) -> None
        self.filename = filename
        self._lock = Lock()
        self.cache_timeout = cache_timeout
        self._servers = set()  # type: Set[Server]
        self._last_cache = 0.0

    def _load_servers(self):
        # type: () -> None
        servers = set()

        with open(self.filename) as servers_file:
            for line in servers_file:
                line = line.split("#", 1)[0].strip()
                if not line:
                    continue

                if line.count(":") != 1:
                    continue

                hostname, port_str = line.split(":")
                try:
                    port = int(port_str)
                except ValueError:
                    continue

                servers.add(Server(hostname, port))

        self._servers = servers

    @property
    def servers(self):
        # type: () -> Set[Server]
        with self._lock:
            now = time.time()
            if self._last_cache + self.cache_timeout > now:
                return self._servers

            self._last_cache = now
            self._load_servers()
            return self._servers

    def __call__(self):
        # type: () -> Set[Server]
        return self.servers



import json
from typing import TYPE_CHECKING

import pytest
from mock import Mock, patch

from groupy.client import Groupy, HTTPClient
from groupy.exc import TimeTravelNotAllowed
from tests.fixtures import user_response  # noqa: F401

if TYPE_CHECKING:
    from typing import Any, Dict, Text


def test_checkpoint(user_response):  # noqa: F811
    # type: (Dict[Text, Any]) -> None
    res1 = Mock()
    res1.body = json.dumps(
        {
            u"checkpoint": 10,
            u"checkpoint_time": 1000,
            u"data": user_response["data"],
            u"status": u"ok",
        }
    )
    res2 = Mock()
    res2.body = json.dumps(
        {
            u"checkpoint": 11,
            u"checkpoint_time": 1001,
            u"data": user_response["data"],
            u"status": u"ok",
        }
    )

    # sunny day
    mock_fetch = Mock()
    mock_fetch.side_effect = [res1, res2]
    with patch.object(HTTPClient, "fetch", mock_fetch):
        client = Groupy(["localhost:8000"])
        client.users.get("oliver@a.co")
        client.users.get("oliver@a.co")

    # time travel not allowed
    mock_fetch.side_effect = [res2, res1]
    with patch.object(HTTPClient, "fetch", mock_fetch), pytest.raises(TimeTravelNotAllowed):
        client = Groupy(["localhost:8000"])
        client.users.get("oliver@a.co")
        client.users.get("oliver@a.co")

    # time travel allowed
    mock_fetch.side_effect = [res2, res1]
    with patch.object(HTTPClient, "fetch", mock_fetch):
        client = Groupy(["localhost:8000"], allow_time_travel=True)
        client.users.get("oliver@a.co")
        client.users.get("oliver@a.co")



import json
from typing import TYPE_CHECKING

from mock import Mock, patch

from groupy.client import Groupy, HTTPClient
from tests.fixtures import permission_response  # noqa: F401
from tests.fixtures import service_account_response  # noqa: F401

if TYPE_CHECKING:
    from typing import Any, Dict, Text


def test_service_account(service_account_response):  # noqa: F811
    # type: (Dict[Text, Any]) -> None
    res = Mock()
    res.body = json.dumps(service_account_response)
    mock_fetch = Mock()
    mock_fetch.side_effect = [res]
    with patch.object(HTTPClient, "fetch", mock_fetch):
        client = Groupy(["localhost:8000"])
        service = client.users.get("service@a.co")
        assert service.enabled
        assert service.groups == {}
        assert service.passwords == []

        assert len(service.permissions) == 1
        assert service.permissions[0].permission == "sudo"
        assert service.permissions[0].argument == "shell"
        assert service.permissions[0].granted_on == 1452796706.894347
        assert service.permissions[0].distance is None
        assert service.permissions[0].path is None

        expected = service_account_response["data"]["user"]["service_account"]
        assert service.service_account == expected


def test_permission(permission_response):  # noqa: F811
    # type: (Dict[Text, Any]) -> None
    res = Mock()
    res.body = json.dumps(permission_response)
    mock_fetch = Mock()
    mock_fetch.side_effect = [res]
    with patch.object(HTTPClient, "fetch", mock_fetch):
        client = Groupy(["localhost:8000"])
        permission = client.permissions.get("grouper.audit.security")
        assert permission.groups == {}
        assert isinstance(permission.audited, bool)
        assert permission.audited is False



#!/usr/bin/env python

import sys

from setuptools import setup

try:
    from typing import List
except Exception:
    pass

# Define __version__.  This is equivalent to execfile but works in Python 3.
with open("groupy/version.py", "r") as version:
    code = compile(version.read(), "groupy/version.py", "exec")
    exec(code)

# Installation requirements.
with open("requirements.txt") as requirements:
    requires = requirements.read().splitlines()

# Test suite requirements.
with open("requirements-dev2.txt") as requirements:
    test_requires = requirements.read().splitlines()

# Add pytest-runner to setup_requires if running setup with the test argument.
setup_requires = []  # type: List[str]
if "test" in sys.argv:
    setup_requires += ["pytest-runner"]

kwargs = {
    "name": "groupy",
    "version": __version__,  # type: ignore  # noqa: F821
    "packages": ["groupy"],
    "description": "Python client library for Grouper",
    "long_description": open("README.rst").read(),
    "author": "Gary M. Josack, Mark Smith, Herbert Ho, Luke Faraone, Russ Allbery",
    "license": "Apache-2.0",
    "install_requires": requires,
    "setup_requires": setup_requires,
    "tests_require": test_requires,
    "url": "https://github.com/dropbox/groupy",
    "classifiers": [
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "Topic :: Software Development",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
}

setup(**kwargs)



import errno
import json
import logging
import socket
from threading import Lock
from typing import NamedTuple, TYPE_CHECKING

from clowncar.backends import Backends
from tornado.httpclient import HTTPClient, HTTPError, HTTPRequest

from groupy import exc
from groupy.collations import Groups, Permissions, ServiceAccounts, Users

try:
    from urllib.parse import urlencode
except ImportError:
    from urllib import urlencode  # type: ignore

if TYPE_CHECKING:
    from clowncar.server import Server
    from typing import Any, Dict, List, Optional

Checkpoint = NamedTuple("Checkpoint", [("checkpoint", int), ("checkpoint_time", float)])


def _checkpoint_is_greater(a, b):
    # type: (Checkpoint, Checkpoint) -> bool
    """Ensure elements of checkpoint 'a' are all greater than or equal to those in
    checkpoint 'b'."""
    return a.checkpoint >= b.checkpoint and a.checkpoint_time >= b.checkpoint_time


class Groupy(object):
    def __init__(
        self,
        servers,  # type: List[Server]
        partition_key=None,  # type: Optional[str]
        timeout=3,  # type: int
        allow_time_travel=False,  # type: bool
        checkpoint=0,  # type: int
        checkpoint_time=0,  # type: float
        mark_bad_timeout=60,  # type: int
        max_backend_tries=5,  # type: int
    ):
        # type: (...) -> None
        """
        The grouper client.

        Args:
            servers (list of clowncar.server.Server): available API servers
            partition_key (str): key to use for picking a server, None defaults
                to hostname
            timeout (int): connection and request sent to tornado's HTTPClient
            allow_time_travel (bool): allow checkpoint[_time] to go backwards
                in subsequent queries
            checkpoint (int): starting checkpoint
            checkpoint_time (float): starting checkpoint unix epoch time
            mark_bad_timeout (int): time in seconds to not use servers that
                have been marked as dead
            max_backend_tries (int): number of backend servers to try before
                giving up and raising a BackendConnectionError
        """

        self._lock = Lock()
        self.timeout = timeout
        self.backends = Backends(servers, partition_key)

        self.checkpoint = Checkpoint(checkpoint, checkpoint_time)

        self.allow_time_travel = allow_time_travel
        self.mark_bad_timeout = mark_bad_timeout
        self.max_backend_tries = max_backend_tries

        self.users = Users(self, "users")
        self.groups = Groups(self, "groups")
        self.permissions = Permissions(self, "permissions")
        self.service_accounts = ServiceAccounts(self, "service_accounts")

    def _try_fetch(self, path, **kwargs):
        # type: (str, **Any) -> Dict[str, Any]
        last_failed_server = None
        for idx in range(self.max_backend_tries):
            try:
                return self._fetch(path, **kwargs)
            except exc.BackendConnectionError as err:
                logging.warning("Marking server {} as dead.".format(err.server.hostname))
                self.backends.mark_dead(err.server, self.mark_bad_timeout)
                last_failed_server = err.server
        raise exc.BackendConnectionError(
            "Tried {} servers, all failed.".format(self.max_backend_tries), last_failed_server
        )

    def _fetch(self, path, **kwargs):
        # type: (str, **Any) -> Dict[str, Any]
        http_client = HTTPClient()
        server = self.backends.server
        url = HTTPRequest(
            "http://{}:{}{}".format(server.hostname, server.port, path),
            connect_timeout=self.timeout,
            request_timeout=self.timeout,
            **kwargs
        )
        try:
            out = json.loads(http_client.fetch(url).body)
        except HTTPError as err:
            message = err.message or ""
            if err.code == 599:
                raise exc.BackendConnectionError(message, server)
            if err.response:
                try:
                    out = json.loads(err.response.body)
                    if "status" not in out:
                        raise exc.BackendIntegrityError(message, server)
                except (ValueError, TypeError):
                    raise exc.BackendIntegrityError(message, server)
            else:
                raise exc.BackendIntegrityError(message, server)
        except socket.error as err:
            if err.errno == errno.ECONNREFUSED:
                raise exc.BackendConnectionError("socket error (Connection Refused)", server)
            raise

        with self._lock:
            new_checkpoint = Checkpoint(out["checkpoint"], out["checkpoint_time"])
            old_checkpoint = self.checkpoint
            if (
                not _checkpoint_is_greater(new_checkpoint, old_checkpoint)
                and not self.allow_time_travel
            ):
                raise exc.TimeTravelNotAllowed(
                    "Received checkpoint of {} when previously {}".format(
                        new_checkpoint, old_checkpoint
                    ),
                    server,
                )
            self.checkpoint = new_checkpoint

        return out

    def authenticate(self, token):
        # type: (str) -> Dict[str, Any]
        return self._try_fetch("/token/validate", method="POST", body=urlencode({"token": token}))



from typing import TYPE_CHECKING

from groupy import exc
from groupy.resources import Group, Permission, ServiceAccount, User

if TYPE_CHECKING:
    from typing import Any, Dict, Iterator, Optional
    from groupy.client import Groupy


class Collection(object):
    def __init__(self, client, name):
        # type: (Groupy, str) -> None
        self.client = client
        self.name = name

    def _get(self, resource=None):
        # type: (Optional[str]) -> Dict[str, Any]
        path = "/{}".format(self.name)
        if resource:
            path += "/{}".format(resource)
        response = self.client._try_fetch(path)
        self._check_error(response)
        return response

    @staticmethod
    def _check_error(response):
        # type: (Dict[str, Any]) -> None
        if response["status"] == "ok":
            return

        combined = []
        for error in response["errors"]:
            code, message = error["code"], error["message"]
            if code == 404:
                raise exc.ResourceNotFound(message)
            combined.append("{}: {}".format(code, message))

        raise exc.ResourceError(", ".join(combined))

    def __iter__(self):
        # type: () -> Iterator[object]
        response = self._get()
        for resource in response["data"][self.name]:
            yield resource


class Groups(Collection):
    def get(self, resource):
        # type: (str) -> Group
        return Group.from_payload(self._get(resource))


class Users(Collection):
    def get(self, resource):
        # type: (str) -> User
        return User.from_payload(self._get(resource))


class ServiceAccounts(Collection):
    def get(self, resource):
        # type: (str) -> User
        """Service accounts do not (yet) have their own meaningful class."""
        return ServiceAccount.from_payload(self._get(resource))


class Permissions(Collection):
    def get(self, resource):
        # type: (str) -> Permission
        return Permission.from_payload(self._get(resource))
